package eu.jonahbauer.wizard.client.libgdx.actors.game.overlay;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.math.Interpolation;
import com.badlogic.gdx.scenes.scene2d.*;
import com.badlogic.gdx.scenes.scene2d.ui.Container;
import com.badlogic.gdx.scenes.scene2d.utils.TextureRegionDrawable;
import eu.jonahbauer.wizard.client.libgdx.UiskinAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.screens.GameScreen;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.*;

public abstract class Overlay extends Action {
    protected static final float OVERLAY_TIME = 5.0f;

    protected final GameScreen screen;
    protected final WizardGame.Data data;
    protected final TextureAtlas atlas;
    private long timeout;

    private Container<?> root;
    private boolean started;
    private boolean finished;

    public Overlay(GameScreen gameScreen, long timeout) {
        this.screen = gameScreen;
        this.data = gameScreen.getData();
        this.atlas = gameScreen.getAtlas();
        this.timeout = timeout;
    }

    @Override
    public boolean act(float delta) {
        if (!started) {
            if (System.currentTimeMillis() > timeout) {
                return true;
            }

            started = true;
            show((Group) getActor());
        }

        if (System.currentTimeMillis() > timeout) {
            finishInternal();
        }

        return finished;
    }

    protected abstract Actor createContent();

    protected Container<?> getRoot() {
        if (root == null) {
            root = new Container<>(createContent());
            root.setBackground(new TextureRegionDrawable(data.uiskinAtlas.findRegion(UiskinAtlas.WHITE)).tint(new Color(0, 0, 0, 0.5f)));
            root.setSize(WizardGame.WIDTH, WizardGame.HEIGHT);
            root.setTouchable(Touchable.enabled);
        }

        return root;
    }

    protected void show(Group parent) {
        parent.addActor(getRoot());
    }

    protected void finishInternal() {
        if (!finished) {
            finished = true;
            getRoot().remove();
        }
    }

    public void finish() {
        getRoot().addAction(sequence(
                targeting(root, alpha(0.0f, .1f, Interpolation.pow2Out)),
                run(this::finishInternal)
        ));
    }

    public void timeout() {
        timeout = 0;
    }
}
