package eu.jonahbauer.wizard.client.libgdx.actors.game.overlay;

import com.badlogic.gdx.math.Interpolation;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.Group;
import com.badlogic.gdx.scenes.scene2d.actions.ParallelAction;
import com.badlogic.gdx.scenes.scene2d.ui.HorizontalGroup;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.scenes.scene2d.ui.VerticalGroup;
import eu.jonahbauer.wizard.client.libgdx.actions.MyActions;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardActor;
import eu.jonahbauer.wizard.client.libgdx.screens.GameScreen;
import eu.jonahbauer.wizard.common.model.Card;

import java.util.Map;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.*;
import static eu.jonahbauer.wizard.client.libgdx.actions.MyActions.*;

public class TrumpOverlay extends Overlay {
    @SuppressWarnings("RedundantTypeArguments")
    private static final Map<Card, Card.Suit> DEFAULT_SUITES = Map.<Card, Card.Suit>ofEntries(
            Map.entry(Card.BLUE_1, Card.Suit.BLUE),
            Map.entry(Card.BLUE_2, Card.Suit.BLUE),
            Map.entry(Card.BLUE_3, Card.Suit.BLUE),
            Map.entry(Card.BLUE_4, Card.Suit.BLUE),
            Map.entry(Card.BLUE_5, Card.Suit.BLUE),
            Map.entry(Card.BLUE_6, Card.Suit.BLUE),
            Map.entry(Card.BLUE_7, Card.Suit.BLUE),
            Map.entry(Card.BLUE_8, Card.Suit.BLUE),
            Map.entry(Card.BLUE_9, Card.Suit.BLUE),
            Map.entry(Card.BLUE_10, Card.Suit.BLUE),
            Map.entry(Card.BLUE_11, Card.Suit.BLUE),
            Map.entry(Card.BLUE_12, Card.Suit.BLUE),
            Map.entry(Card.BLUE_13, Card.Suit.BLUE),
            Map.entry(Card.RED_1, Card.Suit.RED),
            Map.entry(Card.RED_2, Card.Suit.RED),
            Map.entry(Card.RED_3, Card.Suit.RED),
            Map.entry(Card.RED_4, Card.Suit.RED),
            Map.entry(Card.RED_5, Card.Suit.RED),
            Map.entry(Card.RED_6, Card.Suit.RED),
            Map.entry(Card.RED_7, Card.Suit.RED),
            Map.entry(Card.RED_8, Card.Suit.RED),
            Map.entry(Card.RED_9, Card.Suit.RED),
            Map.entry(Card.RED_10, Card.Suit.RED),
            Map.entry(Card.RED_11, Card.Suit.RED),
            Map.entry(Card.RED_12, Card.Suit.RED),
            Map.entry(Card.RED_13, Card.Suit.RED),
            Map.entry(Card.GREEN_1, Card.Suit.GREEN),
            Map.entry(Card.GREEN_2, Card.Suit.GREEN),
            Map.entry(Card.GREEN_3, Card.Suit.GREEN),
            Map.entry(Card.GREEN_4, Card.Suit.GREEN),
            Map.entry(Card.GREEN_5, Card.Suit.GREEN),
            Map.entry(Card.GREEN_6, Card.Suit.GREEN),
            Map.entry(Card.GREEN_7, Card.Suit.GREEN),
            Map.entry(Card.GREEN_8, Card.Suit.GREEN),
            Map.entry(Card.GREEN_9, Card.Suit.GREEN),
            Map.entry(Card.GREEN_10, Card.Suit.GREEN),
            Map.entry(Card.GREEN_11, Card.Suit.GREEN),
            Map.entry(Card.GREEN_12, Card.Suit.GREEN),
            Map.entry(Card.GREEN_13, Card.Suit.GREEN),
            Map.entry(Card.YELLOW_1, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_2, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_3, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_4, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_5, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_6, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_7, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_8, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_9, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_10, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_11, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_12, Card.Suit.YELLOW),
            Map.entry(Card.YELLOW_13, Card.Suit.YELLOW)
    );

    private final String player;
    private final Card card;
    private final Card.Suit suit;

    private final CardActor trumpCardActor;
    private final CardActor trumpSuitActor;

    private boolean animateCard = true;

    public TrumpOverlay(GameScreen gameScreen, String player, Card card, Card.Suit suit) {
        super(gameScreen, Long.MAX_VALUE);
        this.player = player;
        this.card = card;
        this.suit = suit;
        this.trumpCardActor = gameScreen.getTrumpCardActor();
        this.trumpSuitActor = gameScreen.getTrumpSuitActor();
    }

    @Override
    public Actor createContent() {
        var root = new VerticalGroup().columnCenter().space(10f);

        String text;

        if (player == null) {
            text = suit != null ? switch (suit) {
                case YELLOW -> "The trump suit is [#ffff00]yellow[#ffffff].";
                case GREEN -> "The trump suit is [#00ff00]green[#ffffff].";
                case BLUE -> "The trump suit is [#0000ff]blue[#ffffff].";
                case RED -> "The trump suit is [#ff0000]red[#ffffff].";
                default -> "There is no trump suit.";
            } : "To be determined.";
        } else {
            text = player + (suit != null ? switch (suit) {
                case YELLOW -> " chose the trump suit [#ffff00]yellow[#ffffff].";
                case GREEN -> " chose the trump suit [#00ff00]green[#ffffff].";
                case BLUE -> " chose the trump suit [#0000ff]blue[#ffffff].";
                case RED -> " chose the trump suit [#ff0000]red[#ffffff].";
                default -> " has decided there will be no trump suit this round.";
            } : " to be determined.");
        }

        var label = new Label(text, data.skin);
        label.getStyle().font.getData().markupEnabled = true;
        label.setFontScale(1.5f);
        root.addActor(label);

        var cardGroup = new HorizontalGroup().space(20);
        root.addActor(cardGroup);

        if (trumpCardActor.getParent() != null && trumpCardActor.getCard() == card && suit != null) {
            // if card actor is already correct then dont change it
            animateCard = false;
        } else {
            trumpCardActor.remove();
            cardGroup.addActor(trumpCardActor);
            trumpCardActor.setCard(card);
        }

        trumpSuitActor.remove();
        if (suit != DEFAULT_SUITES.get(card)) {
            trumpSuitActor.setRotation(0);
            trumpSuitActor.setOrigin(0, 0);
            cardGroup.addActor(trumpSuitActor);
            trumpSuitActor.setCard(suit);
        }

        return root;
    }

    @Override
    public void show(Group parent) {
        super.show(parent);

        ParallelAction cardAnimation = parallel();
        if (animateCard) {
            cardAnimation.addAction(sequence(
                    targeting(trumpCardActor, removeActorSilently()),
                    targeting(trumpCardActor, changeParent(parent)),
                    targeting(trumpCardActor, moveTo(10, 10, .3f))
            ));
        }

        if (suit != null) {
            cardAnimation.addAction(sequence(
                    targeting(trumpSuitActor, removeActorSilently()),
                    targeting(trumpSuitActor, changeParent(parent)),
                    run(trumpCardActor::toFront),
                    parallel(
                            targeting(trumpSuitActor, rotateTo(-90, .3f)),
                            targeting(trumpSuitActor,
                                      moveTo(10, 10 + (trumpSuitActor.getHeight() + trumpSuitActor.getWidth()) / 2, .3f)
                            )
                    )
            ));
        }

        var root = getRoot();
        root.addAction(sequence(
                delay(OVERLAY_TIME),
                parallel(
                        targeting(root, alpha(0.0f, .3f, Interpolation.pow2Out)),
                        cardAnimation
                ),
                run(this::finishInternal)
        ));
    }

    @Override
    public void finish() {
        MyActions.finish(getRoot().getActions().get(0));
    }
}
