package eu.jonahbauer.wizard.client.libgdx.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Touchable;
import com.badlogic.gdx.scenes.scene2d.ui.Container;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.scenes.scene2d.ui.VerticalGroup;
import com.badlogic.gdx.scenes.scene2d.utils.TextureRegionDrawable;
import com.badlogic.gdx.utils.Align;
import eu.jonahbauer.wizard.client.libgdx.GameAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardActor;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardStack;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardsGroup;
import eu.jonahbauer.wizard.client.libgdx.actors.game.PadOfTruth;
import eu.jonahbauer.wizard.client.libgdx.actors.game.overlay.*;
import eu.jonahbauer.wizard.client.libgdx.state.Game;
import eu.jonahbauer.wizard.client.libgdx.util.Pair;
import eu.jonahbauer.wizard.common.messages.client.InteractionMessage;
import eu.jonahbauer.wizard.common.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.common.messages.player.ContinueMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayCardMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.common.model.Card;
import lombok.Getter;

import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;

import static eu.jonahbauer.wizard.client.libgdx.actions.MyActions.*;
import static eu.jonahbauer.wizard.common.messages.observer.UserInputMessage.Action.*;

public class GameScreen extends MenuScreen {
    @Getter
    private TextureAtlas atlas;

    private Label.LabelStyle labelStyleDefault;
    private Label.LabelStyle labelStyleActive;

    private final Game state;

    private final List<UUID> players;

    private Pair<UUID, UserInputMessage.Action> activePlayer;

    private CardsGroup handCards;
    private CardStack cardStack;
    private PadOfTruth padOfTruth;

    @Getter
    private CardActor trumpCardActor;
    @Getter
    private CardActor trumpSuitActor;

    private VerticalGroup messages;
    private Label persistentMessage;

    private Action currentAction;
    private final Queue<Action> pendingActions = new LinkedList<>();

    private final Map<UUID, Seat> seats = new HashMap<>();
    private final Map<UUID, Label> nameLabels = new HashMap<>();

    private final AtomicBoolean sending = new AtomicBoolean();
    private final AtomicBoolean pendingSync = new AtomicBoolean();

    public GameScreen(WizardGame game) {
        super(game);
        this.state = (Game) game.getClient().getState();
        this.players = new ArrayList<>(state.getPlayers().keySet());
    }

    @Override
    public void show() {
        super.show();
        atlas = new TextureAtlas(Gdx.files.internal(GameAtlas.$PATH));
//        overlay = new TextureRegionDrawable(game.data.uiskinAtlas.findRegion(UiskinAtlas.WHITE)).tint(new Color(0,0,0,0.5f));
        labelStyleDefault = game.data.skin.get(Label.LabelStyle.class);
        labelStyleActive = new Label.LabelStyle(labelStyleDefault);
        labelStyleActive.fontColor = Color.RED;

        seat();
        prepareLabels();

        handCards = new CardsGroup(Collections.emptyList(), atlas);
        handCards.setOnClickListener(card -> send(new PlayCardMessage(card.getCard())));
        var container = new Container<>(handCards);
        container.setPosition(360, 75);
        container.setSize(1200, CardActor.PREF_HEIGHT);
        container.fillX();

        messages = new VerticalGroup().columnCenter().bottom();
        messages.setPosition(360, 85 + CardActor.PREF_HEIGHT);
        messages.setSize(1200, 0);
        messages.setTouchable(Touchable.disabled);

        padOfTruth = new PadOfTruth(game.data.skin, new TextureRegionDrawable(atlas.findRegion(GameAtlas.PAD_OF_TRUTH)));
        padOfTruth.setPosition(1910 - 636f, 10);
        padOfTruth.setOrigin(636f, 0);

        cardStack = new CardStack();
        cardStack.setHoverBounds(0.5f * WizardGame.WIDTH - 50, 0.5f * WizardGame.HEIGHT - 50, 100, 100);

        setNames();

        Gdx.input.setInputProcessor(game.data.stage);
        game.data.stage.addActor(container);
        game.data.stage.addActor(cardStack);
        game.data.stage.addActor(padOfTruth);
        game.data.stage.addActor(messages);
    }

    @Override
    protected void renderBackground(float delta) {
        float scale = Math.max(
                game.data.extendViewport.getWorldWidth() / WizardGame.WIDTH,
                game.data.extendViewport.getWorldHeight() / WizardGame.HEIGHT
        );
        game.batch.setColor(1, 1, 1, 0.25f);
        game.batch.draw(game.data.background, 0,0, scale * WizardGame.WIDTH, scale * WizardGame.HEIGHT);
        game.batch.draw(
                game.data.title,
                (game.data.extendViewport.getWorldWidth() - game.data.title.getRegionWidth() * 0.75f) / 2f,
                (game.data.extendViewport.getWorldHeight() - game.data.title.getRegionHeight() * 0.75f) / 2f,
                0.75f * game.data.title.getRegionWidth(),
                0.75f * game.data.title.getRegionHeight()
        );
    }

    @Override
    protected void renderForeground(float delta) {
        if (pendingActions.size() > 0) {
            var actions = game.data.stage.getRoot().getActions();
            boolean running = false;
            if (currentAction != null) {
                for (int i = 0; i < actions.size; i++) {
                    if (actions.get(i) == currentAction) {
                        running = true;
                        break;
                    }
                }
            }

            if (!running) {
                currentAction = pendingActions.poll();
                game.data.stage.addAction(currentAction);
            }
        } else if (pendingSync.getAndSet(false)) {
            send(new ContinueMessage());
        }
    }

    private void execute(Action action) {
        pendingActions.add(action);
    }

    private void execute(Runnable runnable) {
        pendingActions.add(run(runnable));
    }

    private void seat() {
        var count = players.size();
        Seat[] seats = switch (count) {
            case 3 -> new Seat[] {Seat.TOP_LEFT, Seat.TOP_RIGHT};
            case 4 -> new Seat[] {Seat.LEFT, Seat.TOP, Seat.RIGHT};
            case 5 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP_RIGHT, Seat.RIGHT};
            case 6 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP, Seat.TOP_RIGHT, Seat.RIGHT};
            default -> throw new AssertionError();
        };
        int index = players.indexOf(state.getSelf());
        for (int i = 1; i < count; i++) {
            var player = players.get((index + i) % count);
            var seat = seats[i - 1];
            this.seats.put(player, seat);
        }
        this.seats.put(state.getSelf(), Seat.BOTTOM);
    }

    private void prepareLabels() {
        for (UUID player : players) {
            if (state.getSelf().equals(player)) continue;
            var label = new Label("", game.data.skin);
            var seat = seats.get(player);
            label.setX(seat.getLabelX());
            label.setY(seat.getLabelY());
            label.setAlignment(seat.getLabelAlign());
            this.nameLabels.put(player, label);
            game.data.stage.addActor(label);
        }
    }

    private void setNames() {
        for (int i = 0; i < players.size(); i++) {
            var player = players.get(i);
            var name = state.getPlayers().get(player);
            padOfTruth.setName(i, name);
            if (!state.getSelf().equals(player)) {
                nameLabels.get(player).setText(name);
            }
        }
    }

    public void startRound(int round) {
        execute(parallel(
                run(() -> {
                    if (trumpCardActor != null) {
                        trumpCardActor.remove();
                    }

                    if (trumpSuitActor != null) {
                        trumpSuitActor.remove();
                    }

                    handCards.clearChildren();
                    cardStack.clearChildren();
                }),
                new StartRoundOverlay(this, round)
        ));
    }

    public void startTrick() {
        setActivePlayer(null, null, 0);
        execute(() -> cardStack.clearChildren());
    }

    public void finishTrick(UUID player, List<Card> cards) {
        var seat = seats.get(player);

        var action = parallel();
        execute(sequence(
                run(() -> cardStack.removeAll().forEach(card -> action.addAction(sequence(
                        targeting(card, changeParent(game.data.stage.getRoot())),
                        parallel(
                                targeting(card, rotateTo(0, 0.1f)),
                                targeting(card, moveTo(
                                        seat.getFrontX() - card.getWidth() / 2,
                                        seat.getFrontY() - card.getHeight() / 2,
                                        0.25f
                                ))
                        ),
                        targeting(card, alpha(0, 0.5f)),
                        removeActor(card)
                )))),
                action
        ));
    }

    public void setHand(UUID player, List<Card> cards) {
        if (state.getSelf().equals(player)) {
            execute(() -> handCards.update(cards));
        }
    }

    public void setTrump(Card trumpCard, Card.Suit trumpSuit) {
        if (trumpCardActor == null) {
            trumpCardActor = new CardActor(Card.HIDDEN, atlas);
        }

        if (trumpSuitActor == null) {
            trumpSuitActor = new CardActor(Card.HIDDEN, atlas);
        }

        String player = null;
        if (activePlayer != null && activePlayer.second() == PICK_TRUMP) {
            player = state.getPlayers().get(activePlayer.first());
            setActivePlayer(null, null, 0);
        }

        execute(new TrumpOverlay(this, player, trumpCard, trumpSuit));
    }

    public void addPrediction(int round, UUID player, int prediction) {
        if (activePlayer != null && activePlayer.first().equals(player) && (activePlayer.second() == CHANGE_PREDICTION || activePlayer.second() == MAKE_PREDICTION)) {
            setActivePlayer(null, null, 0);
        }

        addMessage(state.getPlayers().get(player) + " predicted " + prediction + ".");
        execute(() -> padOfTruth.setPrediction(players.indexOf(player), round, prediction));
    }

    public void playCard(UUID player, Card card) {
        if (activePlayer != null && activePlayer.first().equals(player) && activePlayer.second() == PLAY_CARD) {
            setActivePlayer(null, null, 0);
        }

        addMessage(state.getPlayers().get(player) + " played a card.");

        Seat seat = seats.get(player);

        var sequence = sequence();
        sequence.addAction(run(() -> {
            CardActor actor;
            if (state.getSelf().equals(player)) {
                actor = handCards.remove(card);
            } else {
                actor = new CardActor(card, atlas);
                actor.setPosition(seat.getHandX() - actor.getWidth() / 2, seat.getHandY() - actor.getHeight() / 2);
                actor.setRotation(seat.getHandAngle());
            }
            actor.setOrigin(actor.getWidth() / 2, actor.getHeight() / 2);
            cardStack.add(seat, actor);
            sequence.addAction(delay(actor));
            sequence.addAction(delay(0.5f));
        }));
        execute(sequence);
    }

    public void addScores(int round, Map<UUID, Integer> scores) {
        execute(() -> {
            for (int i = 0; i < players.size(); i++) {
                UUID player = players.get(i);
                padOfTruth.setScore(i, round, scores.get(player));
            }
        });
    }

    public void setActivePlayer(UUID player, UserInputMessage.Action action, long timeout) {
        if (activePlayer != null && nameLabels.containsKey(activePlayer.getKey())) {
            var label = nameLabels.get(activePlayer.getKey());
            execute(() -> label.setStyle(labelStyleDefault));
        }

        if (player == null && action == null) {
            activePlayer = null;
            setPersistentMessage(null);
            return;
        }

        activePlayer = Pair.of(player, action);
        if (nameLabels.containsKey(player)) {
            var label = nameLabels.get(player);
            execute(() -> label.setStyle(labelStyleActive));
        }

        if (state.getSelf().equals(player)) {
            setPersistentMessage(null);
            switch (action) {
                case PICK_TRUMP -> execute(new PickTrumpOverlay(this, timeout, false));
                case MAKE_PREDICTION -> execute(new MakePredictionOverlay(this, timeout, state.getRound()));
                case CHANGE_PREDICTION -> execute(new ChangePredictionOverlay(this, timeout, state.getRound(), state.getPredictions().get(state.getSelf())));
                case PLAY_CARD -> setPersistentMessage("It is your turn to play a card.");
            }
            // TODO do something
        } else {
            var name = state.getPlayers().get(player);
            setPersistentMessage("It is " + name + "'s turn to " + switch (action) {
                case CHANGE_PREDICTION -> "change their prediction";
                case JUGGLE_CARD -> "juggle a card";
                case PLAY_CARD -> "play a card";
                case PICK_TRUMP -> "pick the trump suit";
                case MAKE_PREDICTION -> "make a prediction";
                default -> throw new AssertionError();
            });
        }
    }

    public void addMessage(String text) {
        addMessage(text, false);
    }

    public void addMessage(String text, boolean immediate) {
        var label = new Label(text, game.data.skin);
        label.addAction(sequence(
                delay(1.5f),
                alpha(0, 0.25f),
                removeActor()
        ));

        Runnable runnable = () -> {
            if (persistentMessage != null && persistentMessage.getParent() == messages) {
                messages.addActorBefore(persistentMessage, label);
            } else {
                messages.addActor(label);
            }
        };

        if (immediate) {
            runnable.run();
        } else {
            execute(runnable);
        }
    }

    public void setPersistentMessage(String text) {
        execute(() -> {
            if (text != null) {
                if (persistentMessage == null) {
                    persistentMessage = new Label(text, getData().skin);
                } else {
                    persistentMessage.setText(text);
                }
                messages.addActor(persistentMessage);
            } else if (persistentMessage != null) {
                persistentMessage.remove();
            }
        });
    }

    public void send(PlayerMessage message) {
        if (!sending.getAndSet(true) || (message instanceof ContinueMessage)) {
            game.getClient().send(new InteractionMessage(message));
        } else {
            addMessage("Please slow down.", true);
        }
    }

    public void ready(boolean success) {
        if (!pendingSync.get()) {
            sending.set(false);
        }

        if (success && currentAction instanceof Overlay overlay && overlay instanceof InteractionOverlay) {
            overlay.finish();
        }
    }

    public void timeout() {
        addMessage("Timed out.");
        ready(true);
    }

    public void sync() {
        pendingSync.set(true);
        sending.set(true);
    }

    @Override
    public void dispose() {
        if (atlas != null) atlas.dispose();
    }

    public WizardGame.Data getData() {
        return game.data;
    }

    @Getter
    public enum Seat {
        BOTTOM(WizardGame.WIDTH * 0.5f, 0, 0, 0, Align.bottom, WizardGame.WIDTH * 0.5f, 300),
        LEFT(0, WizardGame.HEIGHT * 0.5f, 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomLeft, 117.5f, WizardGame.HEIGHT * 0.5f),
        RIGHT(WizardGame.WIDTH, WizardGame.HEIGHT * 0.5f, WizardGame.WIDTH - 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomRight, WizardGame.WIDTH - 117.5f, WizardGame.HEIGHT * 0.5f),
        TOP_LEFT(WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 200),
        TOP(WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 200),
        TOP_RIGHT(WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 200);

        // position of the hand, should be offscreen
        private final float handX;
        private final float handY;
        private final float handAngle;

        // position of the label
        private final float labelX;
        private final float labelY;
        private final int labelAlign;

        private final float frontX;
        private final float frontY;

        Seat(float handX, float handY, float labelX, float labelY, int labelAlign, float frontX, float frontY) {
            this.handX = handX;
            this.handY = handY;
            this.labelX = labelX;
            this.labelY = labelY;
            this.labelAlign = labelAlign;
            this.frontX = frontX;
            this.frontY = frontY;

            var deltaX = WizardGame.WIDTH * 0.5f - handX;
            var deltaY = WizardGame.HEIGHT * 0.5f - handY;
            this.handAngle = (float) Math.toDegrees(Math.atan2(deltaY, deltaX) + Math.PI / 2);
        }
    }
}
