package eu.jonahbauer.wizard.core.machine.states;

import eu.jonahbauer.wizard.common.messages.observer.TimeoutMessage;
import eu.jonahbauer.wizard.common.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.common.messages.player.ContinueMessage;
import eu.jonahbauer.wizard.common.messages.player.PlayerMessage;
import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.GameState;

import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import static eu.jonahbauer.wizard.common.messages.observer.UserInputMessage.Action.SYNC;
import static eu.jonahbauer.wizard.core.machine.states.GameData.PLAYERS;

public abstract class SyncState extends GameState {
    private final transient Set<UUID> ready = new HashSet<>();

    public SyncState(GameData data) {
        super(data);
    }

    @Override
    public Optional<GameState> onEnter(Game game) {
        game.notify(new UserInputMessage(null, SYNC, getSyncTimeout(game, true)));
        return syncTimeout(game);
    }

    @Override
    public Optional<GameState> onMessage(Game game, UUID player, PlayerMessage message) {
        if (message instanceof ContinueMessage) {
            ready.add(player);

            if (ready.size() == get(PLAYERS).size()) {
                return Optional.of(getNextState());
            } else {
                return Optional.empty();
            }
        } else {
            return super.onMessage(game, player, message);
        }
    }

    @Override
    public Optional<GameState> onTimeout(Game game) {
        game.notify(new TimeoutMessage());
        return Optional.of(getNextState());
    }

    protected abstract GameState getNextState();
}
